//-----------------------------------------------------------------------
// Copyright (c) 2018 Snap Inc.
//-----------------------------------------------------------------------
#if defined(CMDLINE)
#include "metal_std.metal"
#include "metalstd_vs.metal"
#include "metalstd_fs.metal"
#else
#include <metal_std.metal>
#include <metalstd_vs.metal>
#include <metalstd_fs.metal>
#endif

constant float MAIN_PASS = 0.0;
constant float COLOR_PASS = 0.1;
constant float SHADOW_PASS = 0.2;
constant float OUTLINE_PASS = 0.3;
constant float BACKGROUND_PASS = 0.4;

struct TextVertex
{
    float3 position [[attribute(0)]];
    float2 texture0 [[attribute(1)]];
    float2 texture1 [[attribute(2)]];
    float passIdentifierAttr [[attribute(3)]];
};

struct TextVaryings {
    float4 position [[position]];
    half2 texture0;
    half2 texture1;
    half passIdentifier;
};

struct TextUniforms
{
    float4 mainColor;
    float3x3 mainFillTextureTransform;
    
    float4 shadowColor;
    float3x3 shadowFillTextureTransform;
    
    float4 outlineColor;
    float3x3 outlineFillTextureTransform;
    
    float4 backgroundColor;
    float3x3 backgroundFillTextureTransform;
};

constant bool MAIN_FILL_COLOR [[function_constant(FUNCONST_VS_USER_BIAS+0)]];
constant bool MAIN_FILL_TEXTURE [[function_constant(FUNCONST_VS_USER_BIAS+1)]];
constant bool ENABLE_SHADOW [[function_constant(FUNCONST_VS_USER_BIAS+2)]];
constant bool SHADOW_FILL_TEXTURE [[function_constant(FUNCONST_VS_USER_BIAS+3)]];
constant bool SHADOW_FILL_COLOR [[function_constant(FUNCONST_VS_USER_BIAS+4)]];
constant bool ENABLE_OUTLINE [[function_constant(FUNCONST_VS_USER_BIAS+5)]];
constant bool OUTLINE_FILL_TEXTURE [[function_constant(FUNCONST_VS_USER_BIAS+6)]];
constant bool OUTLINE_FILL_COLOR [[function_constant(FUNCONST_VS_USER_BIAS+7)]];
constant bool ENABLE_BACKGROUND [[function_constant(FUNCONST_VS_USER_BIAS+8)]];
constant bool BACKGROUND_FILL_TEXTURE [[function_constant(FUNCONST_VS_USER_BIAS+9)]];
constant bool BACKGROUND_FILL_COLOR [[function_constant(FUNCONST_VS_USER_BIAS+10)]];



bool isPass(float pass, float identifier) {
    if((identifier > (pass - 0.05)) && (identifier < (pass + 0.05))) {
        return true;
    }
    return false;
}

VSMAIN_WITH_VERTEXTYPE_RETTYPE_AND_UNIFORMTYPE(TextVertex, TextVaryings, TextUniforms)
{
    Vertex vtx;
    vtx.position = customVertex.position;
    vtx.texture0 = customVertex.texture0;
    vtx.texture1 = customVertex.texture1;
    sc_Vertex_t v = sc_LoadVertexAttributes(vtx);
    
    sc_ProcessVertexNoReturn(v);
    
    TextVaryings out;
    out.passIdentifier = customVertex.passIdentifierAttr;
    out.position = varyings.position;
    
    out.texture0 = half2(customVertex.texture0);
    out.texture1 = half2(customVertex.texture1);
    return out;
}

// fragment shader function
fragment half4 fs_main(TextVaryings in [[stage_in]],
                       constant TextUniforms& uniforms [[ buffer(2) ]],
                       texture2d<half> mainTexture,
                       sampler mainTextureSampler,
                       texture2d<half> colorTexture,
                       sampler colorTextureSampler,
                       DECLARETEXTUREOPTIONAL(mainFillTexture,mainFillTextureSampler,MAIN_FILL_TEXTURE),
                       DECLARETEXTUREOPTIONAL(shadowFillTexture,shadowFillTextureSampler,SHADOW_FILL_TEXTURE),
                       DECLARETEXTUREOPTIONAL(outlineFillTexture,outlineFillTextureSampler,OUTLINE_FILL_TEXTURE),
                       DECLARETEXTUREOPTIONAL(backgroundFillTexture,backgroundFillTextureSampler,BACKGROUND_FILL_TEXTURE))
{
    half4 fillColor = half4(1.0);
    
    float3 fillUvs = float3(float(fract(in.texture1.x)), float(fract(in.texture1.y)), 1.0);
        
    if (isPass(MAIN_PASS, in.passIdentifier) || isPass(COLOR_PASS, in.passIdentifier)) {
        if (MAIN_FILL_COLOR) {
            fillColor = half4(uniforms.mainColor);
        } else if (MAIN_FILL_TEXTURE) {
            float2 texture1 = (uniforms.mainFillTextureTransform * fillUvs).xy;
            texture1.y = 1.0 - texture1.y;
            fillColor = mainFillTexture.sample(mainFillTextureSampler, texture1);
        } else {
            assert(false);
        }
    } else if (isPass(SHADOW_PASS, in.passIdentifier)) {
        if (SHADOW_FILL_COLOR) {
            fillColor = half4(uniforms.shadowColor);
        } else if (SHADOW_FILL_TEXTURE) {
            float2 texture1 = (uniforms.shadowFillTextureTransform * fillUvs).xy;
            texture1.y = 1.0 - texture1.y;
            fillColor = shadowFillTexture.sample(shadowFillTextureSampler, texture1);
        } else {
            assert(false);
        }
    } else if (isPass(OUTLINE_PASS, in.passIdentifier)) {
        if (OUTLINE_FILL_COLOR) {
            fillColor = half4(uniforms.outlineColor);
        } else if (OUTLINE_FILL_TEXTURE) {
            float2 texture1 = (uniforms.outlineFillTextureTransform * fillUvs).xy;
            texture1.y = 1.0 - texture1.y;
            fillColor = outlineFillTexture.sample(outlineFillTextureSampler, texture1);
        } else {
            assert(false);
        }
    } else if (isPass(BACKGROUND_PASS, in.passIdentifier)) {
        if (BACKGROUND_FILL_COLOR) {
            return half4(uniforms.backgroundColor);
        } else if (BACKGROUND_FILL_TEXTURE) {
            float2 texture1 = (uniforms.backgroundFillTextureTransform * fillUvs).xy;
            texture1.y = 1.0 - texture1.y;
            return backgroundFillTexture.sample(backgroundFillTextureSampler, texture1);
        } else {
            assert(false);
        }
    } else {
        assert(false);
    }
    
    if (isPass(COLOR_PASS, in.passIdentifier)) {
        half4 colorSample = colorTexture.sample(colorTextureSampler, float2(in.texture0));
        return half4(colorSample.rgb, colorSample.a * fillColor.a);
    }
    
    half alphaSample = mainTexture.sample(mainTextureSampler, float2(in.texture0)).r;
    return half4(fillColor.rgb, alphaSample * fillColor.a);
};

